/* 
 * tkCanvMeter.c --
 *
 *	This file implements horizontal and  vertical meter and
 *      circular gauge items for canvas widgets.
 *
 */

#include <stdio.h>
#include "tk.h"
#include "tkInt.h"
#include "tkPort.h"

/*
 * The structure below defines the record for each meter item.
 */

typedef struct MeterItem  {
    Tk_Item header;		/* Generic stuff that's the same for all
				 * types.  MUST BE FIRST IN STRUCTURE. */
    double bbox[4];		/* Coordinates of bounding box for rectangle
				 * or oval (x1, y1, x2, y2).  Item includes
				 * x1 and x2 but not y1 and y2. */
    int width;			/* Width of outline. */
    int type;			/* 0 = hmeter, 1 = vmeter, 2 = gauge */
    double percentage;		/* percentage full */
    XColor *outlineColor;	/* Color for outline. */
    XColor *fillColor;		/* Color for filling meter/gauge. */
    XColor *unfillColor;	/* Color for filling meter/gauge. */
    GC outlineGC;		/* Graphics context for outline. */
    GC fillGC;			/* Graphics context for filling item. */
    GC unfillGC;		/* Graphics context for unfilled portion. */
} MeterItem;

/*
 * Information used for parsing configuration specs:
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_COLOR, "-fill", (char *) NULL, (char *) NULL,
	"red", Tk_Offset(MeterItem, fillColor), 0},
    {TK_CONFIG_COLOR, "-outline", (char *) NULL, (char *) NULL,
	"black", Tk_Offset(MeterItem, outlineColor), 0},
    {TK_CONFIG_COLOR, "-unfill", (char *) NULL, (char *) NULL,
	"white", Tk_Offset(MeterItem, unfillColor), 0},
    {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
	(char *) NULL, 0, TK_CONFIG_NULL_OK, &tk_CanvasTagsOption},
    {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
	"1", Tk_Offset(MeterItem, width), TK_CONFIG_DONT_SET_DEFAULT},
    {TK_CONFIG_DOUBLE, "-percentage", (char *) NULL, (char *) NULL,
	"0", Tk_Offset(MeterItem, percentage), 0},
    {TK_CONFIG_INT, "-type", (char *) NULL, (char *) NULL,
	"0", Tk_Offset(MeterItem, type), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Prototypes for procedures defined in this file:
 */

static void		ComputeMeterBbox _ANSI_ARGS_((Tk_Canvas canvas,
			    MeterItem *meterPtr));
static int		ConfigureMeter _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
			    char **argv, int flags));
static int		CreateMeter _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, struct Tk_Item *itemPtr,
			    int argc, char **argv));
static void		DeleteMeter _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, Display *display));
static void		DisplayMeter _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, Display *display, Drawable dst,
			    int x, int y, int width, int height));
static int		MeterCoords _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
			    char **argv));
static int		MeterToPostscript _ANSI_ARGS_((Tcl_Interp *interp,
			    Tk_Canvas canvas, Tk_Item *itemPtr, int prepass));
static int		MeterToArea _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double *areaPtr));
static double		MeterToPoint _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double *pointPtr));
static void		ScaleMeter _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double originX, double originY,
			    double scaleX, double scaleY));
static void		TranslateMeter _ANSI_ARGS_((Tk_Canvas canvas,
			    Tk_Item *itemPtr, double deltaX, double deltaY));

/*
 * The structures below defines the rectangle and oval item types
 * by means of procedures that can be invoked by generic item code.
 */

Tk_ItemType tkMeterType = {
    "meter",				/* name */
    sizeof(MeterItem),			/* itemSize */
    CreateMeter,			/* createProc */
    configSpecs,			/* configSpecs */
    ConfigureMeter,			/* configureProc */
    MeterCoords,			/* coordProc */
    DeleteMeter,			/* deleteProc */
    DisplayMeter,			/* displayProc */
    0,					/* alwaysRedraw */
    MeterToPoint,			/* pointProc */
    MeterToArea,			/* areaProc */
    MeterToPostscript,			/* postscriptProc */
    ScaleMeter,				/* scaleProc */
    TranslateMeter,			/* translateProc */
    (Tk_ItemIndexProc *) NULL,		/* indexProc */
    (Tk_ItemCursorProc *) NULL,		/* icursorProc */
    (Tk_ItemSelectionProc *) NULL,	/* selectionProc */
    (Tk_ItemInsertProc *) NULL,		/* insertProc */
    (Tk_ItemDCharsProc *) NULL,		/* dTextProc */
    (Tk_ItemType *) NULL		/* nextPtr */
};


/*
 *--------------------------------------------------------------
 *
 * CreateMeter --
 *
 *	This procedure is invoked to create a new meter
 *	item in a canvas.
 *
 * Results:
 *	A standard Tcl return value.  If an error occurred in
 *	creating the item, then an error message is left in
 *	interp->result;  in this case itemPtr is left uninitialized,
 *	so it can be safely freed by the caller.
 *
 * Side effects:
 *	A new meter item is created.
 *
 *--------------------------------------------------------------
 */

static int
CreateMeter(Tcl_Interp* interp, Tk_Canvas canvas, Tk_Item* itemPtr, int argc,
char *argv[] )
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;

    if (argc < 4) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		Tk_PathName(Tk_CanvasTkwin(canvas)), "\" create ",
		itemPtr->typePtr->name, " x1 y1 x2 y2 ?options?",
		(char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Carry out initialization that is needed in order to clean
     * up after errors during the the remainder of this procedure.
     */

    meterPtr->width = 1;
    meterPtr->type = 0;
    meterPtr->outlineColor = NULL;
    meterPtr->fillColor = NULL;
    meterPtr->unfillColor = NULL;
    meterPtr->outlineGC = None;
    meterPtr->fillGC = None;
    meterPtr->unfillGC = None;

    /*
     * Process the arguments to fill in the item record.
     */

    if ((Tk_CanvasGetCoord(interp, canvas, argv[0],
		&meterPtr->bbox[0]) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[1],
		&meterPtr->bbox[1]) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[2],
		    &meterPtr->bbox[2]) != TCL_OK)
	    || (Tk_CanvasGetCoord(interp, canvas, argv[3],
		    &meterPtr->bbox[3]) != TCL_OK)) {
	return TCL_ERROR;
    }

    if (ConfigureMeter(interp, canvas, itemPtr, argc-4, argv+4, 0)
	    != TCL_OK) {
	DeleteMeter(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas)));
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * MeterCoords --
 *
 *	This procedure is invoked to process the "coords" widget
 *	command on rectangles and ovals.  See the user documentation
 *	for details on what it does.
 *
 * Results:
 *	Returns TCL_OK or TCL_ERROR, and sets interp->result.
 *
 * Side effects:
 *	The coordinates for the given item may be changed.
 *
 *--------------------------------------------------------------
 */

static int
MeterCoords(Tcl_Interp* interp, Tk_Canvas canvas, Tk_Item* itemPtr, int argc,
char *argv[] )
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    char c0[TCL_DOUBLE_SPACE], c1[TCL_DOUBLE_SPACE];
    char c2[TCL_DOUBLE_SPACE], c3[TCL_DOUBLE_SPACE];

    if (argc == 0) {
	Tcl_PrintDouble(interp, meterPtr->bbox[0], c0);
	Tcl_PrintDouble(interp, meterPtr->bbox[1], c1);
	Tcl_PrintDouble(interp, meterPtr->bbox[2], c2);
	Tcl_PrintDouble(interp, meterPtr->bbox[3], c3);
	Tcl_AppendResult(interp, c0, " ", c1, " ", c2, " ", c3,
		(char *) NULL);
    } else if (argc == 4) {
	if ((Tk_CanvasGetCoord(interp, canvas, argv[0],
		    &meterPtr->bbox[0]) != TCL_OK)
		|| (Tk_CanvasGetCoord(interp, canvas, argv[1],
		    &meterPtr->bbox[1]) != TCL_OK)
		|| (Tk_CanvasGetCoord(interp, canvas, argv[2],
			&meterPtr->bbox[2]) != TCL_OK)
		|| (Tk_CanvasGetCoord(interp, canvas, argv[3],
			&meterPtr->bbox[3]) != TCL_OK)) {
	    return TCL_ERROR;
	}
	ComputeMeterBbox(canvas, meterPtr);
    } else {
	sprintf(interp->result,
		"wrong # coordinates:  expected 0 or 4, got %d",
		argc);
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ConfigureMeter --
 *
 *	This procedure is invoked to configure various aspects
 *	of a meter item, such as its border and background colors.
 *
 * Results:
 *	A standard Tcl result code.  If an error occurs, then
 *	an error message is left in interp->result.
 *
 * Side effects:
 *	Configuration information, such as colors and stipple
 *	patterns, may be set for itemPtr.
 *
 *--------------------------------------------------------------
 */

static int
ConfigureMeter(Tcl_Interp* interp, Tk_Canvas canvas, Tk_Item* itemPtr,
int argc, char* argv[], int flags )
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    XGCValues gcValues;
    GC newGC;
    unsigned long mask;
    Tk_Window tkwin;

    tkwin = Tk_CanvasTkwin(canvas);
    if (Tk_ConfigureWidget(interp, tkwin, configSpecs, argc, argv,
	    (char *) meterPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * A few of the options require additional processing, such as
     * graphics contexts.
     */

    if (meterPtr->width < 1) {
	meterPtr->width = 1;
    }

    if (meterPtr->percentage < 0)
	meterPtr->percentage = 0;
    else if (meterPtr->percentage > 1)
	meterPtr->percentage = 1;

    if (meterPtr->outlineColor == NULL) {
	newGC = None;
    } else {
	gcValues.foreground = meterPtr->outlineColor->pixel;
	gcValues.cap_style = CapProjecting;
	gcValues.line_width = meterPtr->width;
	mask = GCForeground|GCCapStyle|GCLineWidth;
	newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (meterPtr->outlineGC != None) {
	Tk_FreeGC(Tk_Display(tkwin), meterPtr->outlineGC);
    }
    meterPtr->outlineGC = newGC;

    if (meterPtr->fillColor == NULL) {
	newGC = None;
    } else {
	gcValues.foreground = meterPtr->fillColor->pixel;
	mask = GCForeground;
	newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (meterPtr->fillGC != None) {
	Tk_FreeGC(Tk_Display(tkwin), meterPtr->fillGC);
    }
    meterPtr->fillGC = newGC;

    if (meterPtr->unfillColor == NULL) {
	newGC = None;
    } else {
	gcValues.foreground = meterPtr->unfillColor->pixel;
	gcValues.cap_style = CapProjecting;
	gcValues.line_width = meterPtr->width;
	mask = GCForeground|GCCapStyle|GCLineWidth;
	newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    if (meterPtr->unfillGC != None) {
	Tk_FreeGC(Tk_Display(tkwin), meterPtr->unfillGC);
    }
    meterPtr->unfillGC = newGC;

    ComputeMeterBbox(canvas, meterPtr);

    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteMeter --
 *
 *	This procedure is called to clean up the data structure
 *	associated with a meter item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Resources associated with itemPtr are released.
 *
 *--------------------------------------------------------------
 */

static void
DeleteMeter(Tk_Canvas canvas, Tk_Item* itemPtr, Display* display)
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;

    if (meterPtr->outlineColor != NULL) {
	Tk_FreeColor(meterPtr->outlineColor);
    }
    if (meterPtr->fillColor != NULL) {
	Tk_FreeColor(meterPtr->fillColor);
    }
    if (meterPtr->unfillColor != NULL) {
	Tk_FreeColor(meterPtr->unfillColor);
    }
    if (meterPtr->outlineGC != None) {
	Tk_FreeGC(display, meterPtr->outlineGC);
    }
    if (meterPtr->fillGC != None) {
	Tk_FreeGC(display, meterPtr->fillGC);
    }
    if (meterPtr->unfillGC != None) {
	Tk_FreeGC(display, meterPtr->unfillGC);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ComputeMeterBbox --
 *
 *	This procedure is invoked to compute the bounding box of
 *	all the pixels that may be drawn as part of a meter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The fields x1, y1, x2, and y2 are updated in the header
 *	for itemPtr.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static void
ComputeMeterBbox(
    Tk_Canvas canvas,			/* Canvas that contains item. */
    MeterItem *meterPtr )		/* Item whose bbox is to be
					 * recomputed. */
{
    int bloat;

    /*
     * Make sure that the first coordinates are the lowest ones.
     */

    if (meterPtr->bbox[1] > meterPtr->bbox[3]) {
	double tmp;
	tmp = meterPtr->bbox[3];
	meterPtr->bbox[3] = meterPtr->bbox[1];
	meterPtr->bbox[1] = tmp;
    }
    if (meterPtr->bbox[0] > meterPtr->bbox[2]) {
	double tmp;
	tmp = meterPtr->bbox[2];
	meterPtr->bbox[2] = meterPtr->bbox[0];
	meterPtr->bbox[0] = tmp;
    }

    if (meterPtr->outlineColor == NULL) {
	bloat = 0;
    } else {
	bloat = (meterPtr->width+1)/2;
    }
    meterPtr->header.x1 = (int)(meterPtr->bbox[0] - bloat);
    meterPtr->header.y1 = (int)(meterPtr->bbox[1] - bloat);
    meterPtr->header.x2 = (int)(meterPtr->bbox[2] + bloat + 1);
    meterPtr->header.y2 = (int)(meterPtr->bbox[3] + bloat + 1);
}

/*
 *--------------------------------------------------------------
 *
 * DisplayMeter --
 *
 *	This procedure is invoked to draw a meter item in a
 *      given drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	ItemPtr is drawn in drawable using the transformation
 *	information in canvas.
 *
 *--------------------------------------------------------------
 */

static void
DisplayMeter(
    Tk_Canvas canvas,			/* Canvas that contains item. */
    Tk_Item *itemPtr,			/* Item to be displayed. */
    Display *display,			/* Display on which to draw item. */
    Drawable drawable,			/* Pixmap or window in which to draw */
					/* item. */
    int x, int y,			/* Describes region of canvas that */
    int width, int height )		/* must be redisplayed (not used). */
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    short x1, y1, x2, y2, z;

    /*
     * Compute the screen coordinates of the bounding box for the item.
     * Make sure that the bbox is at least one pixel large, since some
     * X servers will die if it isn't.
     */

    Tk_CanvasDrawableCoords(canvas, meterPtr->bbox[0], meterPtr->bbox[1],
	    &x1, &y1);
    Tk_CanvasDrawableCoords(canvas, meterPtr->bbox[2], meterPtr->bbox[3],
	    &x2, &y2);
    if (x2 <= x1) x2 = x1+1;
    if (y2 <= y1) y2 = y1+1;

    switch( meterPtr->type ) {
       case 1:	/* vmeter */
          z = (int)(y1 + (y2 - y1) * (1.0 - meterPtr->percentage) + 0.5);
          if (meterPtr->unfillGC != None)
             XFillRectangle(display, drawable, meterPtr->unfillGC,
                x1, y1, (unsigned int) (x2-x1), (unsigned int) (z-y1));
          if (meterPtr->fillGC != None)
             XFillRectangle(display, drawable, meterPtr->fillGC,
                x1, z, (unsigned int) (x2-x1), (unsigned int) (y2-z));
          if (meterPtr->outlineGC != None)
             XDrawRectangle(display, drawable, meterPtr->outlineGC,
                x1, y1, (unsigned) (x2-x1), (unsigned) (y2-y1));
       break;
       case 2:	/* gauge */
          if (meterPtr->unfillGC != None)
             XFillArc(display, drawable, meterPtr->unfillGC,
                (int)x1, (int)y1, (unsigned) (x2-x1), (unsigned) (y2-y1),
                (int)(360 * meterPtr->percentage * 64), (int)360*64);
          if (meterPtr->fillGC != None)
             XFillArc(display, drawable, meterPtr->fillGC,
                (int)x1, (int)y1, (unsigned) (x2-x1), (unsigned) (y2-y1),
                (int)0, (int)(360 * meterPtr->percentage * 64));
          if (meterPtr->outlineGC != None)
             XDrawArc(display, drawable, meterPtr->outlineGC,
                (int)x1, (int)y1, (unsigned) (x2-x1), (unsigned) (y2-y1),
                (int)0, (int)360*64);
       break;
       case 0:	/* hmeter */
       default:
          z = (short)(x1 + (x2 - x1) * meterPtr->percentage + 0.5);
          if (meterPtr->unfillGC != None)
             XFillRectangle(display, drawable, meterPtr->unfillGC,
                z, y1, (unsigned int) (x2-z), (unsigned int) (y2-y1));
          if (meterPtr->fillGC != None)
             XFillRectangle(display, drawable, meterPtr->fillGC,
                x1, y1, (unsigned int) (z-x1), (unsigned int) (y2-y1));
          if (meterPtr->outlineGC != None)
             XDrawRectangle(display, drawable, meterPtr->outlineGC,
                x1, y1, (unsigned) (x2-x1), (unsigned) (y2-y1));
       break;
    }
}

/*
 *--------------------------------------------------------------
 *
 * MeterToPoint --
 *
 *	Computes the distance from a given point to a given
 *	meter, in canvas units.
 *
 * Results:
 *	The return value is 0 if the point whose x and y coordinates
 *	are coordPtr[0] and coordPtr[1] is inside the rectangle.  If the
 *	point isn't inside the rectangle then the return value is the
 *	distance from the point to the rectangle.  If itemPtr is filled,
 *	then anywhere in the interior is considered "inside"; if
 *	itemPtr isn't filled, then "inside" means only the area
 *	occupied by the outline.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static double
MeterToPoint(
    Tk_Canvas canvas,		/* Canvas containing item. */
    Tk_Item *itemPtr,		/* Item to check against point. */
    double *pointPtr )		/* Pointer to x and y coordinates. */
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    double xDiff, yDiff, x1, y1, x2, y2, inc, tmp;

    /*
     * Generate a new larger rectangle that includes the border
     * width, if there is one.
     */

    x1 = meterPtr->bbox[0];
    y1 = meterPtr->bbox[1];
    x2 = meterPtr->bbox[2];
    y2 = meterPtr->bbox[3];
    if (meterPtr->outlineGC != None) {
	inc = meterPtr->width/2.0;
	x1 -= inc;
	y1 -= inc;
	x2 += inc;
	y2 += inc;
    }

    /*
     * If the point is inside the rectangle, handle specially:
     * distance is 0 if rectangle is filled, otherwise compute
     * distance to nearest edge of rectangle and subtract width
     * of edge.
     */

    if ((pointPtr[0] >= x1) && (pointPtr[0] < x2)
		&& (pointPtr[1] >= y1) && (pointPtr[1] < y2)) {
	if ((meterPtr->fillGC != None) || (meterPtr->outlineGC == None)) {
	    return 0.0;
	}
	xDiff = pointPtr[0] - x1;
	tmp = x2 - pointPtr[0];
	if (tmp < xDiff) {
	    xDiff = tmp;
	}
	yDiff = pointPtr[1] - y1;
	tmp = y2 - pointPtr[1];
	if (tmp < yDiff) {
	    yDiff = tmp;
	}
	if (yDiff < xDiff) {
	    xDiff = yDiff;
	}
	xDiff -= meterPtr->width;
	if (xDiff < 0.0) {
	    return 0.0;
	}
	return xDiff;
    }

    /*
     * Point is outside rectangle.
     */

    if (pointPtr[0] < x1) {
	xDiff = x1 - pointPtr[0];
    } else if (pointPtr[0] > x2)  {
	xDiff = pointPtr[0] - x2;
    } else {
	xDiff = 0;
    }

    if (pointPtr[1] < y1) {
	yDiff = y1 - pointPtr[1];
    } else if (pointPtr[1] > y2)  {
	yDiff = pointPtr[1] - y2;
    } else {
	yDiff = 0;
    }

    return hypot(xDiff, yDiff);
}

/*
 *--------------------------------------------------------------
 *
 * MeterToArea --
 *
 *	This procedure is called to determine whether an item
 *	lies entirely inside, entirely outside, or overlapping
 *	a given meter.
 *
 * Results:
 *	-1 is returned if the item is entirely outside the area
 *	given by meterPtr, 0 if it overlaps, and 1 if it is entirely
 *	inside the given area.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static int
MeterToArea(
    Tk_Canvas canvas,		/* Canvas containing item. */
    Tk_Item *itemPtr,		/* Item to check against rectangle. */
    double *areaPtr )		/* Pointer to array of four coordinates
				 * (x1, y1, x2, y2) describing rectangular
				 * area.  */
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    double halfWidth;

    halfWidth = meterPtr->width/2.0;
    if (meterPtr->outlineGC == None) {
	halfWidth = 0.0;
    }

    if ((areaPtr[2] <= (meterPtr->bbox[0] - halfWidth))
	    || (areaPtr[0] >= (meterPtr->bbox[2] + halfWidth))
	    || (areaPtr[3] <= (meterPtr->bbox[1] - halfWidth))
	    || (areaPtr[1] >= (meterPtr->bbox[3] + halfWidth))) {
	return -1;
    }
    if ((meterPtr->fillGC == None) && (meterPtr->outlineGC != None)
	    && (areaPtr[0] >= (meterPtr->bbox[0] + halfWidth))
	    && (areaPtr[1] >= (meterPtr->bbox[1] + halfWidth))
	    && (areaPtr[2] <= (meterPtr->bbox[2] - halfWidth))
	    && (areaPtr[3] <= (meterPtr->bbox[3] - halfWidth))) {
	return -1;
    }
    if ((areaPtr[0] <= (meterPtr->bbox[0] - halfWidth))
	    && (areaPtr[1] <= (meterPtr->bbox[1] - halfWidth))
	    && (areaPtr[2] >= (meterPtr->bbox[2] + halfWidth))
	    && (areaPtr[3] >= (meterPtr->bbox[3] + halfWidth))) {
	return 1;
    }
    return 0;
}

/*
 *--------------------------------------------------------------
 *
 * ScaleMeter --
 *
 *	This procedure is invoked to rescale a meter item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The meter referred to by temPtr is rescaled
 *	so that the following transformation is applied to all
 *	point coordinates:
 *		x' = originX + scaleX*(x-originX)
 *		y' = originY + scaleY*(y-originY)
 *
 *--------------------------------------------------------------
 */

static void
ScaleMeter(
    Tk_Canvas canvas,			/* Canvas containing rectangle. */
    Tk_Item *itemPtr,			/* Rectangle to be scaled. */
    double originX, double originY,	/* Origin about which to scale rect. */
    double scaleX,			/* Amount to scale in X direction. */
    double scaleY )			/* Amount to scale in Y direction. */
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;

    meterPtr->bbox[0] = originX + scaleX*(meterPtr->bbox[0] - originX);
    meterPtr->bbox[1] = originY + scaleY*(meterPtr->bbox[1] - originY);
    meterPtr->bbox[2] = originX + scaleX*(meterPtr->bbox[2] - originX);
    meterPtr->bbox[3] = originY + scaleY*(meterPtr->bbox[3] - originY);
    ComputeMeterBbox(canvas, meterPtr);
}

/*
 *--------------------------------------------------------------
 *
 * TranslateMeter --
 *
 *	This procedure is called to move a meter by a given amount.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The position of the meter is offset by
 *	(xDelta, yDelta), and the bounding box is updated in the
 *	generic part of the item structure.
 *
 *--------------------------------------------------------------
 */

static void
TranslateMeter(
    Tk_Canvas canvas,			/* Canvas containing item. */
    Tk_Item *itemPtr,			/* Item that is being moved. */
    double deltaX, double deltaY )	/* Amount by which item is to be
					 * moved. */
{
    MeterItem *meterPtr = (MeterItem *) itemPtr;

    meterPtr->bbox[0] += deltaX;
    meterPtr->bbox[1] += deltaY;
    meterPtr->bbox[2] += deltaX;
    meterPtr->bbox[3] += deltaY;
    ComputeMeterBbox(canvas, meterPtr);
}

/*
 *--------------------------------------------------------------
 *
 * MeterToPostscript --
 *
 *	This procedure is called to generate Postscript for
 *	meter items.
 *
 * Results:
 *	The return value is a standard Tcl result.  If an error
 *	occurs in generating Postscript then an error message is
 *	left in interp->result, replacing whatever used to be there.
 *	If no error occurs, then Postscript for the meter is
 *	appended to the result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
MeterToPostscript(
    Tcl_Interp *interp,			/* Interpreter for error reporting. */
    Tk_Canvas canvas,			/* Information about overall canvas. */
    Tk_Item *itemPtr,			/* Item for which Postscript is
					 * wanted. */
    int prepass	)			/* 1 means this is a prepass to
					 * collect font information;  0 means
					 * final Postscript is being created. */
{
    char fillPathCmd[500], unfillPathCmd[500], outlinePathCmd[500], string[100];
    MeterItem *meterPtr = (MeterItem *) itemPtr;
    double x1, x2, y1, y2;

    const char* mFmt = "%.15g %.15g moveto %.15g 0 rlineto 0 %.15g rlineto %.15g 0 rlineto closepath\n";

    const char* gFmt = "matrix currentmatrix\n%.15g %.15g translate %.15g %.15g scale %d 0 moveto 0 0 1 %.15g %.15g arc closepath\nsetmatrix\n";

    x1 = meterPtr->bbox[0];
    x2 = meterPtr->bbox[2];
    y1 = Tk_CanvasPsY(canvas, meterPtr->bbox[1]);
    y2 = Tk_CanvasPsY(canvas, meterPtr->bbox[3]);

    /*
     * Generate a string that creates 3 paths for the meter.
     * This is the only part of the procedure's code that is type-
     * specific.
     */

    if (meterPtr->type == 0) {		/* hmeter */
       double z = x1 + (x2 - x1) * meterPtr->percentage;
       sprintf(fillPathCmd,    mFmt, x1, y1,  z-x1, y2-y1, x1- z );
       sprintf(unfillPathCmd,  mFmt,  z, y1, x2- z, y2-y1, z -x2 );
       sprintf(outlinePathCmd, mFmt, x1, y1, x2-x1, y2-y1, x1-x2 );
    }
    else if (meterPtr->type == 1) {	/* vmeter */
       double z = y1 + (y2 - y1) * (1.0 - meterPtr->percentage);
       sprintf(fillPathCmd,    mFmt, x1,  z, x2-x1, y2- z, x1-x2 );
       sprintf(unfillPathCmd,  mFmt, x1, y1, x2-x1, z -y1, x1-x2 );
       sprintf(outlinePathCmd, mFmt, x1, y1, x2-x1, y2-y1, x1-x2 );
    }
    else {				/* gauge */
	sprintf(fillPathCmd, gFmt,
           (x1 + x2)/2, (y1 + y2)/2, (x2 - x1)/2, (y1 - y2)/2,
           0, 0.0, 360.0 * meterPtr->percentage );
	sprintf(unfillPathCmd, gFmt,
           (x1 + x2)/2, (y1 + y2)/2, (x2 - x1)/2, (y1 - y2)/2,
           0, 360.0 * meterPtr->percentage, 360.0);
	sprintf(outlinePathCmd, gFmt,
           (x1 + x2)/2, (y1 + y2)/2, (x2 - x1)/2, (y1 - y2)/2,
           1, 0.0, 360.0);
    }

    /*
     * First draw the filled area of the meter.
     */

    if (meterPtr->fillColor != NULL) {
	Tcl_AppendResult(interp, fillPathCmd, (char *) NULL);
	if (Tk_CanvasPsColor(interp, canvas, meterPtr->fillColor)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	Tcl_AppendResult(interp, "fill\n", (char *) NULL);
    }

    /*
     * Next, draw the unfilled area of the meter.
     */

    if (meterPtr->unfillColor != NULL) {
	Tcl_AppendResult(interp, unfillPathCmd, (char *) NULL);
	if (Tk_CanvasPsColor(interp, canvas, meterPtr->unfillColor)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	Tcl_AppendResult(interp, "fill\n", (char *) NULL);
    }

    /*
     * Now draw the outline.
     */

    if (meterPtr->outlineColor != NULL) {
	Tcl_AppendResult(interp, outlinePathCmd, (char *) NULL);
	sprintf(string, "%d setlinewidth", meterPtr->width);
	Tcl_AppendResult(interp, string,
		" 0 setlinejoin 2 setlinecap\n", (char *) NULL);
	if (Tk_CanvasPsColor(interp, canvas, meterPtr->outlineColor)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	Tcl_AppendResult(interp, "stroke\n", (char *) NULL);
    }
    return TCL_OK;
}
